#!/usr/bin/env python
# -*- coding: utf8 -*-
"""
contactmap: ContactMap class for protein contact maps and related functions.

Revision 3-April-2014
"""

import sys #stderr, stdout
import needle

class ContactMap(object):
    """Protein contact map.
    
    Members:
    contacts: a set of tuples (i,j) representing 1-based residue
              positions, where i<=j.
    seq: a string representing the sequence of the protein
              represented by the contact map
    metadata: other data concerning the origin of the contact map
              stored as a {string: string} dictionary
    """

    def __init__(self, sequence, contact_list=None, meta=None):
        """Construct a new ContactMap object
        
        Args:
            sequence (str): protein sequence
            contact_list: list of tuples (int,int) representing contacts
            meta: optional {string: string} dictionary of metadata
                  Keys should be the names of the data
                  Values should be the data values
                  e.g. "cutoff": "8.0"
                       "pdb": "1ek0"
        Returns:
            Nothing
        Raises:
            Nothing
        """
        if sequence == None:
            sequence = ""
        self.seq = str(sequence)
        
        if contact_list == None:
            contact_list = []
        self.contacts = set()
        for pair in contact_list:
            self.add_contact(pair)
        
        if meta == None:
            meta = {}
        self.metadata = meta.copy()
        
    def add_contact(self, pair):
        """Add a contact between a pair of residues to the map.
        
        Args:
            pair: tuple (int, int) representing pair of residue positions
                in the protein
        Returns:
            Nothing
        Raises:
            CMError if a contact is added referencing residue positions
            outside the range of the sequence e.g. residue positions less
            than one, or greater than the length of the sequence
            CMError if the contact is supplied in a bad format, i.e. not
            a tuple of length 2
        """
        pair = (int(pair[0]), int(pair[1]))
        if len(pair) != 2:
            raise CMError("bad contact; not in format (i,j)")
        if max(pair) > len(self.seq) or min(pair) < 1:
            raise CMError("residue position outside range of sequence" +
                          str(pair))
        if pair[0] > pair[1]:
            self.contacts.add((pair[1], pair[0]))
        else:
            self.contacts.add(pair)
            
    def remove_contact(self, pair):
        """Removes a contact between a pair of residues from the map.
        
        Args:
            pair -- tuple (int, int) representing pair of residue positions
                    in the protein
        Returns:
            Nothing
        Raises:
            Nothing
        """
        try:
            if pair[0] > pair[1]:
                self.contacts.remove((pair[1], pair[0]))
            else:
                self.contacts.remove(pair)
        except KeyError:
            sys.stderr.write("Warning: contact " + str(pair) +
                             " not found in contact map\n")
            #Uncomment to change warning to error:
            #raise KeyError("contact (" + str(pair) + ") not found")
            
    def get_contact_list(self):
        """Get a list of tuples representing contacts present in the map.
        
        Args:
            None
        Returns:
            A list of (int, int) tuples. The list is sorted by the value of
            the first member of the tuple (increasing).
        Raises:
            Nothing
        """
        return sorted(list(self.contacts))
    
    def remove_neighbours(self, threshold):
        """Remove residues along the diagonal from the contact map."""
        #Doesn't assume pair is actually of length 2
        #Just to be sure everything is int:
        threshold = int(threshold)
        for pair in self.get_contact_list():
            if abs(pair[0] - pair[1]) <= threshold:
                self.remove_contact((pair[0], pair[1]))
        
def compare(cm1, cm2, verbose=True, names=("Contact Map 1", "Contact Map 2")):
    """Compare cm1 to cm2
    
    Args:
        cm1, cm2: ContactMap objects to compare
        verbose: switch to print messages to stdout
        names: tuple of names of contact maps for display
    Returns:
        a tuple (true_positives, false_positives, false_negatives)
        where:
        true_positives is the number of common contacts,
        false_positives is the number of contacts only present in cm1, and
        false_negatives is the number of contacts only present in cm2
    """
    s1_aligned, s2_aligned = needle.needle(cm1.seq, cm2.seq, needle.blosum62,
                                         semi=True, debug=False)
    if verbose:
        print "Alignment:"
        needle.pretty_print(s1_aligned, s2_aligned)
    
    #Create mapping from positions in cm1 to positions in cm2
    #based on alignment
    convert_dic = {}
    for i in range(1, len(s1_aligned)+1):
        #what number a.a. is at this column in alignment?
        position1 = i - s1_aligned[:i].count('-')
        position2 = i - s2_aligned[:i].count('-')
        if s1_aligned[i-1] != '-':
            convert_dic[position1] = position2
            
    converted_contacts = set()
    for pair in cm1.contacts:
        converted_contacts.add((convert_dic[pair[0]], convert_dic[pair[1]])) 
    
    true_positives = len([p for p in converted_contacts if p in cm2.contacts])
    false_positives = len(cm1.contacts) - true_positives
    false_negatives = len(cm2.contacts) - true_positives
    
    if verbose:
        print "%6s  %30s | %30s" % ("", names[0], names[1])
        print "%6s: %30s | %30s" % ("Length", len(cm1.seq), len(cm2.seq))
        print "%6s: %30s | %30s" % ("Common", true_positives, true_positives)
        print "%6s: %30s | %30s" % ("Unique", false_positives, false_negatives)
        print "%6s: %30s | %30s" % ("Total", len(cm1.contacts), 
                                  len(cm2.contacts))
    return (true_positives, false_positives, false_negatives)
    
def read(filename):
    """Parse a contact map file (CMView format) and return a ContactMap object.
    
    Args:
        filename -- path to contact map file to parse
    Returns:
        A ContactMap object representing the contents of the input file.
    Raises:
        IOError if file can't be opened.
        CMError if input file is badly formatted / cannot be parsed.
    """
    cm = None
    with open(filename, 'r') as infile:
        for line in infile:
            if "#SEQUENCE" in line:
                sequence = line.split()[1]
                cm = ContactMap(sequence, None)
            elif "#" in line:
                pass
            else:
                (word1, word2) = line.split()[:2]
                try:
                    cm.add_contact((word1, word2))
                except NameError:
                    raise CMError("bad input file; cannot parse")
    if cm is None:
        raise CMError('error parsing input file %s' % filename)
    return cm
    
def read_psicov(filename, sequence, cutoff=float('-inf')):
    """Parse a PSICOV output file and return a ContactMap object.
    
    Args:
        filename: name of input file to parse
        sequence: sequence of protein represented in the contact map
        cutoff:   cutoff score--pick only contacts with scores greater than
                  or equal to this value
    Returns:
        a ContactMap object
    Raises:
        nothing
    """
    metadata = {"Cutoff score": str(cutoff), "Source file": filename}
    cm = ContactMap(sequence, None, meta=metadata)
    with open(filename, 'r') as infile:
        for line in infile:
            words = line.split()
            #first two columns in file are residue numbers, fifth column
            #contains score
            if float(words[4]) >= cutoff:
                cm.add_contact((int(words[0]), int(words[1])))
    return cm
    
def read_ev(filename, sequence, cutoff=float('-inf'), offset=0):
    """Parse an EVfold output file and return a ContactMap object.
    
    Args:
        filename: name of input file to parse
        sequence: sequence of protein represented in the contact map
    Returns:
        a ContactMap object
    Raises:
        nothing
    
    """
    metadata = {"Cutoff score": str(cutoff), "Source file": filename}
    cm = ContactMap(sequence, None, meta=metadata)
    with open(filename, 'r') as infile:
        for line in infile:
            words = line.split()
            #first and third columns in file are residue numbers, sixth column
            #contains score
            if float(words[5]) >= cutoff:
                cm.add_contact((int(words[0])-offset, int(words[2])-offset))
    return cm
        
def write(cm, filename=None):
    """ (ContactMap) -> None
    Write a ContactMap object to stdout/file.
    
    Writes a file in CMView format. Header lines start with "#".
    Two columns specify pairs of residue-residue contacts.
    
    Args:
        cm: ContactMap object
        filename: path of file to write
                  if no filename specified, use stdout
    Returns:
        Nothing
    Raises:
        Nothing
    """
    if filename == None:
        outfile = sys.stdout
    else:
        outfile = open(filename, 'w')
    
    outfile.write("#CMVIEW GRAPH FILE ver: 1.0\n")
    outfile.write("#SEQUENCE: %s\n" % cm.seq)
    for key in cm.metadata:
        outfile.write("#%s: %s\n" % (key, cm.metadata[key]))
    for pair in cm.get_contact_list():
        outfile.write("%i\t%i\n" % (pair[0], pair[1]))
    if outfile != sys.stdout: #Don't close stdout and create problems elsewhere
        outfile.close()
    
class CMError(Exception):
    """Class for exceptions specific to the ContactMap class."""
    pass
    
